# -*- coding: iso-8859-1 -*-
"""
    MoinMoin - ImprovedTableParser library

    @copyright: 2012 by Paul Boddie <paul@boddie.org.uk>
    @license: GNU GPL (v2 or later), see COPYING.txt for details.
"""

from MoinMoin import wikiutil
from shlex import shlex
from StringIO import StringIO
from MoinSupport import *
import re

# Regular expressions.

syntax = {
    # For section markers.
    "markers"   : (r"^\s*(?P<n>\\+)(?P<b>{|})(?P=n)(?P=b)(?P=n)(?P=b)", re.MULTILINE),
    "marker"    : (r"(\\+)",                                            0),

    # At start of line:
    "sections"  : (r"(^\s*{{{.*?^\s*}}})",  re.MULTILINE | re.DOTALL),  # {{{ ... }}}
    "rows"      : (r"^==",                  re.MULTILINE),              # ==

    # Within text:
    "columns"   : (r"\|\|[ \t]*",           0),                         # || ws-excl-nl

    # At start of column text:
    "column"    : (r"^\s*<(.*?)>\s*(.*)",   re.DOTALL),                 # ws < attributes > ws
    }

patterns = {}
for name, (value, flags) in syntax.items():
    patterns[name] = re.compile(value, re.UNICODE | flags)

# Other regular expressions.

leading_number_regexp = re.compile(r"\d*")

# Constants.

up_arrow = u'\u2191'
down_arrow = u'\u2193'

# Functions.

def parse(s):

    "Parse 's', returning a table definition."

    s = replaceMarkers(s)

    table_attrs = {}
    rows = []

    # The following will be redefined upon the construction of the first row.

    row_attrs = {}
    columns = []

    # Process exposed text and sections.

    exposed = True

    # Initially, start a new row.

    row_continued = False

    for region in patterns["sections"].split(s):

        # Only look for table features in exposed text.

        if exposed:

            # Extract each row from the definition.

            for row_text in patterns["rows"].split(region):

                # Only create a new row when a boundary has been found.

                if not row_continued:
                    if columns:
                        extractAttributes(columns[0][0], row_attrs, table_attrs)

                    row_attrs = {}
                    columns = []
                    rows.append((row_attrs, columns))
                    column_continued = False

                # Extract each column from the row.

                for text in patterns["columns"].split(row_text):

                    # Only create a new column when a boundary has been found.

                    if not column_continued:

                        # Extract the attribute and text sections.

                        match = patterns["column"].search(text)
                        if match:
                            attribute_text, text = match.groups()
                            columns.append([parseAttributes(attribute_text, True), text])
                        else:
                            columns.append([{}, text])

                    else:
                        columns[-1][1] += text

                    # Permit columns immediately following this one.

                    column_continued = False

                # Permit a continuation of the current column.

                column_continued = True

                # Permit rows immediately following this one.

                row_continued = False

            # Permit a continuation if the current row.

            row_continued = True

        # Write any section into the current column.

        else:
            columns[-1][1] += region

        exposed = not exposed

    if columns:
        extractAttributes(columns[0][0], row_attrs, table_attrs)

    return table_attrs, rows

def extractAttributes(attrs, row_attrs, table_attrs):

    """
    Extract row- and table-level attributes from 'attrs', storing them in
    'row_attrs' and 'table_attrs' respectively.
    """

    for name, value in attrs.items():
        if name.startswith("row") and name != "rowspan":
            row_attrs[name] = value
            del attrs[name]
        elif name.startswith("table"):
            table_attrs[name] = value
            del attrs[name]

def replaceMarkers(s):

    "Convert the section notation in 's'."

    l = []
    last = 0

    # Get each marker and convert it.

    for match in patterns["markers"].finditer(s):
        start, stop = match.span()
        l.append(s[last:start])

        # Convert the marker.

        marker = []
        brace = True
        for text in patterns["marker"].split(match.group()):
            if brace:
                marker.append(text)
            else:
                marker.append(text[:-1])
            brace = not brace

        l.append("".join(marker))
        last = stop
    else:
        l.append(s[last:])

    return "".join(l)

def parseAttributes(s, escape=True):

    """
    Parse the table attributes string 's', returning a mapping of names to
    values. If 'escape' is set to a true value, the attributes will be suitable
    for use with the formatter API. If 'escape' is set to a false value, the
    attributes will have any quoting removed.
    """

    attrs = {}
    f = StringIO(s)
    name = None
    need_value = False

    for token in shlex(f):

        # Capture the name if needed.

        if name is None:
            name = escape and wikiutil.escape(token) or strip_token(token)

        # Detect either an equals sign or another name.

        elif not need_value:
            if token == "=":
                need_value = True
            else:
                attrs[name.lower()] = escape and "true" or True
                name = wikiutil.escape(token)

        # Otherwise, capture a value.

        else:
            # Quoting of attributes done similarly to wikiutil.parseAttributes.

            if token:
                if escape:
                    if token[0] in ("'", '"'):
                        token = wikiutil.escape(token)
                    else:
                        token = '"%s"' % wikiutil.escape(token, 1)
                else:
                    token = strip_token(token)
    
            attrs[name.lower()] = token
            name = None
            need_value = False

    return attrs

def strip_token(token):

    "Return the given 'token' stripped of quoting."

    if token[0] in ("'", '"') and token[-1] == token[0]:
        return token[1:-1]
    else:
        return token

# Formatting of embedded content.
# NOTE: Borrowed from EventAggregator.

def getParserClass(request, format):

    """
    Return a parser class using the 'request' for the given 'format', returning
    a plain text parser if no parser can be found for the specified 'format'.
    """

    try:
        return wikiutil.searchAndImportPlugin(request.cfg, "parser", format or "plain")
    except wikiutil.PluginMissingError:
        return wikiutil.searchAndImportPlugin(request.cfg, "parser", "plain")

def getFormatterClass(request, format):

    """
    Return a formatter class using the 'request' for the given output 'format',
    returning a plain text formatter if no formatter can be found for the
    specified 'format'.
    """

    try:
        return wikiutil.searchAndImportPlugin(request.cfg, "formatter", format or "text_plain")
    except wikiutil.PluginMissingError:
        return wikiutil.searchAndImportPlugin(request.cfg, "formatter", "text_plain")

def formatText(text, request, fmt):

    "Format the given 'text' using the specified 'request' and formatter 'fmt'."

    parser_cls = getParserClass(request, request.page.pi["format"])
    parser = parser_cls(text, request, line_anchors=False)
    return request.redirectedOutput(parser.format, fmt, inhibit_p=True)

# Sorting utilities.

def get_sort_columns(s, start=0):

    """
    Split the comma-separated string 's', extracting the column specifications
    of the form <column>["n"] where the prefix "n" indicates an optional
    numeric conversion for that column. Column indexes start from the specified
    'start' value (defaulting to 0).
    """

    sort_columns = []
    for column_spec in s.split(","):
        column_spec = column_spec.strip()

        ascending = True
        if column_spec.endswith("d"):
            column_spec = column_spec[:-1]
            ascending = False

        # Extract the conversion indicator and column index.
        # Ignore badly-specified columns.

        try:
            column = get_number(column_spec)
            suffix = column_spec[len(column):]
            fn = converters[suffix]
            sort_columns.append((max(0, int(column) - start), fn, ascending))
        except ValueError:
            pass

    return sort_columns

def get_column_types(sort_columns):

    """
    Return a dictionary mapping column indexes to conversion functions.
    """

    d = {}
    for column, fn, ascending in sort_columns:
        d[column] = fn, ascending
    return d

def get_number(s):

    "From 's', get any leading number."

    match = leading_number_regexp.match(s)
    if match:
        return match.group()
    else:
        return ""

def to_number(s, request):

    """
    Convert 's' to a number, discarding any non-numeric trailing data.
    Return an empty string if 's' is empty.
    """

    if s:
        return int(get_number(s))
    else:
        return s

def to_plain_text(s, request):

    "Convert 's' to plain text."

    fmt = getFormatterClass(request, "text_plain")(request)
    fmt.page = request.page
    return formatText(s, request, fmt)

converters = {
    "n" : to_number,
    "" : to_plain_text,
    }

suffixes = {}
for key, value in converters.items():
    suffixes[value] = key

class Sorter:

    "A sorting helper class."

    def __init__(self, sort_columns, request):
        self.sort_columns = sort_columns
        self.request = request

    def __call__(self, row1, row2):
        row_attrs1, columns1 = row1
        row_attrs2, columns2 = row2

        # Apply the conversions to each column, comparing the results.

        for column, fn, ascending in self.sort_columns:
            column_attrs1, text1 = columns1[column]
            column_attrs2, text2 = columns2[column]

            # Ignore a column when a conversion is not possible.

            try:
                value1 = fn(text1, self.request)
                value2 = fn(text2, self.request)

                # Avoid empty strings appearing earlier than other values.

                if value1 == "" and value2 != "":
                    result = 1
                elif value1 != "" and value2 == "":
                    result = -1
                else:
                    result = cmp(value1, value2)

                # Where the columns differ, return a result observing the sense
                # (ascending or descending) of the comparison for the column.

                if result != 0:
                    return ascending and result or -result

            except ValueError:
                pass

        return 0

def write_sort_control(request, columnnumber, fmt, write, sort_columns, column_types, columns, table_name, start=0):

    """
    Write a sort control in a pop-up element which provides a list of links
    corresponding to modified sort criteria.
    """

    _ = request.getText

    write(fmt.div(1, css_class="sortcolumns"))

    # Start with the existing criteria without this column being involved.

    revised_sort_columns = [(column, fn, ascending)
        for (column, fn, ascending) in sort_columns if column != columnnumber]

    # Get the specification of this column.

    columnfn, columnascending = column_types.get(columnnumber, (to_plain_text, True))
    newsortcolumn = columnnumber, columnfn, columnascending
    newsortcolumn_reverse = columnnumber, columnfn, not columnascending
    newlabel = columns[columnnumber][1].strip()

    # Show this column in all possible places in the sorting criteria.

    write(fmt.number_list(1))

    just_had_this_column = False

    for i, (column, fn, ascending) in enumerate(sort_columns):
        new_sort_columns = revised_sort_columns[:]
        new_sort_columns.insert(i, newsortcolumn)
        label = columns[column][1].strip()

        arrow = columnascending and down_arrow or up_arrow
        arrow_reverse = not columnascending and down_arrow or up_arrow

        sortcolumns = get_sort_column_output(new_sort_columns)
        new_sort_columns[i] = newsortcolumn_reverse
        sortcolumns_reverse = get_sort_column_output(new_sort_columns)

        # Columns permitting the insertion of the selected column.

        if column != columnnumber and not just_had_this_column:
            write(fmt.listitem(1, css_class="sortcolumn"))

            # Pop-up element showing the column inserted before the sort column.

            write(fmt.span(1, css_class="sortcolumn-container"))
            write(fmt.span(1, css_class="newsortcolumn"))
            write(formatText(newlabel, request, fmt))

            write_sort_link(write, request, fmt, table_name, sortcolumns, arrow, "sortdirection")
            write_sort_link(write, request, fmt, table_name, sortcolumns_reverse, arrow_reverse, "sortdirection")

            write(fmt.span(0))
            write(fmt.span(0))

            # Link for selection of the modified sort criteria.

            write_sort_link(write, request, fmt, table_name, sortcolumns, u"%s %s" % (label, arrow), "")

        # Columns permitting removal or modification.

        else:
            write(fmt.listitem(1))

            # Either show the column without a link, since the column to be
            # inserted is already before the current column.

            if just_had_this_column:
                just_had_this_column = False
                write(fmt.span(1, css_class="unlinkedcolumn"))
                write(formatText(label, request, fmt))
                write(fmt.span(0))

            # Or show the column with a link for its removal.

            else:
                just_had_this_column = True
                sortcolumns_revised = get_sort_column_output(revised_sort_columns)
                write_sort_link(write, request, fmt, table_name, sortcolumns_revised, u"%s %s" % (label, arrow), "removecolumn")

            # Alternative sort direction.

            write_sort_link(write, request, fmt, table_name, sortcolumns_reverse, arrow_reverse, "altdirection")

        write(fmt.listitem(0))

    if not just_had_this_column:

        # Write the sorting criteria with this column at the end.

        new_sort_columns = revised_sort_columns[:]
        new_sort_columns.append(newsortcolumn)

        sortcolumns = get_sort_column_output(new_sort_columns)
        new_sort_columns[-1] = newsortcolumn_reverse
        sortcolumns_reverse = get_sort_column_output(new_sort_columns)

        arrow = columnascending and down_arrow or up_arrow
        arrow_reverse = not columnascending and down_arrow or up_arrow

        write(fmt.listitem(1, css_class="sortcolumn", style="list-style-type: none"))

        # Pop-up element showing the column inserted before the sort column.

        write(fmt.span(1, css_class="sortcolumn-container"))
        write(fmt.span(1, css_class="newsortcolumn"))
        write(formatText(newlabel, request, fmt))

        write_sort_link(write, request, fmt, table_name, sortcolumns, arrow, "sortdirection")
        write_sort_link(write, request, fmt, table_name, sortcolumns_reverse, arrow_reverse, "sortdirection")

        write(fmt.span(0))
        write(fmt.span(0))

        write_sort_link(write, request, fmt, table_name, sortcolumns, _("..."), "")

        write(fmt.listitem(0))

    write(fmt.number_list(0))

    write(fmt.div(0))

def write_sort_link(write, request, fmt, table_name, sortcolumns, label, css_class):

    "Write a link expressing sort criteria."

    write(fmt.url(1, "?%s#%s" % (
        wikiutil.makeQueryString("%s-sortcolumns=%s" % (table_name, sortcolumns)),
        fmt.qualify_id(fmt.sanitize_to_id(table_name))
        ), css_class=css_class))
    write(formatText(label, request, fmt))
    write(fmt.url(0))

def get_sort_column_output(columns, start=0):

    "Return the output criteria for the given 'columns' indexed from 'start'."

    return ",".join([("%d%s%s" % (column + start, suffixes[fn], not ascending and "d" or ""))
        for (column, fn, ascending) in columns])

# Common formatting functions.

def formatTable(text, request, fmt, attrs=None):

    """
    Format the given 'text' using the specified 'request' and formatter 'fmt'.
    The optional 'attrs' can be used to control the presentation of the table.
    """

    # Parse the table region.

    table_attrs, table = parse(text)

    # Define the table name and an anchor attribute.

    table_name = attrs.get("name")
    if table_name:
        table_attrs["tableid"] = table_name
    else:
        table_name = table_attrs.get("tableid")

    # Get the underlying column types.

    column_types = get_column_types(get_sort_columns(attrs.get("columntypes", "")))

    # Get sorting criteria from the region.

    region_sortcolumns = attrs.get("sortcolumns", "")

    # Update the column types from the sort criteria.

    column_types.update(get_column_types(get_sort_columns(region_sortcolumns)))

    # Determine the applicable sort criteria using the request.

    if table_name:
        sortcolumns = getQualifiedParameter(request, table_name, "sortcolumns")
    else:
        sortcolumns = None

    if sortcolumns is None:
        sortcolumns = region_sortcolumns

    # Define the final sort criteria.

    sort_columns = get_sort_columns(sortcolumns)

    # Update the column types from the final sort criteria.

    column_types.update(get_column_types(sort_columns))

    # Sort the rows according to the values in each of the specified columns.

    data_start = int(table_name and getQualifiedParameter(request, table_name, "headers") or attrs.get("headers", "1"))

    if sort_columns:
        headers = table[:data_start]
        data = table[data_start:]

        # Perform the sort and reconstruct the table.

        sorter = Sorter(sort_columns, request)
        data.sort(cmp=sorter)
        table = headers + data

    # Permit sorting because column types for sorting may have been defined.

    else:
        sort_columns = []

    # Write the table.

    write = request.write
    write(fmt.table(1, table_attrs))

    for rownumber, (row_attrs, columns) in enumerate(table):
        write(fmt.table_row(1, row_attrs))
        sortable = column_types and rownumber == data_start - 1

        for columnnumber, (column_attrs, column_text) in enumerate(columns):
            write(fmt.table_cell(1, column_attrs))

            if sortable:
                write(fmt.div(1, css_class="sortablecolumn"))

            write(formatText(column_text, request, fmt))

            # Add sorting controls, if appropriate.

            if sortable:
                write_sort_control(request, columnnumber, fmt, write, sort_columns, column_types, columns, table_name)
                write(fmt.div(0))

            write(fmt.table_cell(0))

        write(fmt.table_row(0))

    write(fmt.table(0))

# vim: tabstop=4 expandtab shiftwidth=4
